/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * afdp.h - Adaptive File Distribution Protocol
 *
 * rcsid[] = "$Id: afdp.h,v 1.52 1998/07/02 19:37:36 steve Exp $";
 *
 */

#include "patchlevel.h"

#include <limits.h>		/* this defines _POSIX_PATH_MAX usually      */
#include <sys/types.h>		/* this defines time_t, u_short, u_long etc. */

/* first define a bunch of variable types, to hide all details together */
typedef u_short port_t;
typedef u_long  host_t;
typedef short family_t;
typedef u_short msg_id_t;
typedef u_long  group_id_t;
typedef u_long seq_t;
typedef u_char data_t;
#define boolean_t u_char

/* now, define all the different packet types in the system */
/* UDP packet types */
#define MSG_SEND 0			/* client => group */
#define MSG_RESEND 1			/* client => group */
#define MSG_SKIP 2			/* client => group */
#define FIND_SECRETARY_REQUEST 3	/* client or sub => secretary */
#define FIND_SECRETARY_REPLY 4		/* secretary -> client or sub */

/* TCP packet types */
#define JOIN_REQUEST 5		/* subscriber -> secretary  */
#define JOIN_REPLY 6		/* secretary -> subscriber  */
#define LEAVE_REQUEST 7		/* subscriber -> secretary  */
#define LEAVE_REPLY 8		/* secretary -> subscriber  */
#define SHUTDOWN 9		/* secretary -> subscriber  */

#define PUBLISH_REQUEST 10	/* client -> secretary  */
#define PUBLISH_PERMISSION 11	/* secretary -> client  */
#define PUBLISH_COMPLETE 12	/* client -> secretary  */
#define SLOWDOWN_REQUEST 13	/* sub or sec -> client */
#define RESEND_REQUEST 14	/* subscriber -> client */

/* miscellaneous defines */
#define AFDP_SEARCH_PORT 6123		/* for UDP broadcasts to secretary */
#define AFDP_RESERVED_PORT 223		/* ditto, using reserved ports     */
#define NULLHOST 0			/* empty host field in memberslist */
#define SUCCESS	1			/* call was successful		   */
#define FAIL	0			/* call was not successful	   */

#define MODE_BROADCAST		0
#define MODE_UNICAST		1
#define MODE_MULTICAST		2

#define OK		0
#define BUSY		1
#define DONE		2
#define REJECT		3

/* packet structure definitions */

/* used for FIND_SECRETARY_REQUEST & FIND_SECRETARY_REPLY */
typedef struct find_secretary_packet {
	u_char packet_type;
	u_char authdata[16];            /* Authentication data              */
	group_id_t group_id;
	port_t admin_port;	/* will be used by subscribers */
	port_t publish_port;	/* will be used by clients     */
	boolean_t responsible;	/* able to become secretary    */
} find_secretary_packet_t;

/* used for JOIN_REQUEST, JOIN_REPLY, LEAVE_REQUEST, LEAVE_REPLY & SHUTDOWN */
typedef struct admin_packet {
	u_char packet_type;	
	u_char authdata[16];            /* Authentication data              */
	group_id_t group_id;
	port_t group_data_port;
	time_t time;			/* pseudo-timestamp */
	u_char capability;		/* do we support multicast? */
	u_char status;			/* SUCCESS or FAIL */
} admin_packet_t;

/* used for PUBLISH_REQUEST, PUBLISH_PERMISSION & PUBLISH_COMPLETE */
typedef struct publish_packet {
	u_char packet_type;	
	u_char authdata[16];            /* Authentication data              */
	group_id_t group_id;
	u_long msg_size;
	msg_id_t msg_id;		/* actual msg_id assigned   */
	port_t group_data_port;		/* NOTE: change in port */
	u_char status;			/* OK/BUSY/FAIL */
	time_t time;			/* time taken by client to publish  */
					/* or time to wait before retry     */
					/* or permission timestamp          */
	u_char publish_mode;		/* unicast, multicast, or broadcast */
	time_t interval;		/* initial inter-packet delay	    */
	host_t multicast_address;	/* only used if multicasting msgs   */
	u_char multicast_ttl;		/* only used if multicasting msgs   */
	u_short host_list_size;		/* length of host id list */
} publish_packet_t;
/* the above packet will be followed by the list of hosts in the group */
/* like this: host_t group_list[host_list_size]; */
	
/* used for SLOWDOWN_REQUEST & RESEND_REQUEST & SHUTDOWN */
typedef struct control_packet {		/* can be issued by sub & sec */
	u_char packet_type;	
	u_char authdata[16];            /* Authentication data              */
	msg_id_t msg_id;
	seq_t seq_number_low;
	seq_t seq_number_high;
} control_packet_t;

/*
 * the maximum size of the data portion of ethernet frame is 1500 bytes
 *
 * by experimentation, we found that the maximum size of a UDP datagram
 * is 9000 bytes. since this is greater than the size of an ethernet frame,
 * the IP layer must fragment and reassemble it.
 *
 * a broadcast UDP datagram cannot be fragmented
 * since the IP header needs 20 bytes, and the UDP header needs 8 bytes
 * this leaves 1472 bytes for user-data
 *
 * part of our user-data field (DATA_PACKET_NONDATA_SIZE bytes) is taken
 * up by non-data, the rest is message data.
 * 
 * if unicasting to subsribers:
 * we will fill in MAX_UNICAST_DATA_SIZE = (9000 - DATA_PACKET_NONDATA_SIZE)
 *
 * if broadcasting to subsribers:
 * we will fill in 1472 - DATA_PACKET_NONDATA_SIZE with message data
 */

#define UDP_SIZE 9000			/* max size of UDP datagram */

#define DATA_PACKET_NONDATA_SIZE (sizeof(group_id_t) + sizeof(msg_id_t) + \
	sizeof(port_t) + sizeof(time_t) + 2 * sizeof(seq_t) + \
	2 * sizeof(u_long) + sizeof(u_short) + 18 * sizeof(u_char))

/*
 * the magic number in the next line should be 1472, but for some reason
 * we need to use 1468 on MIPS machines, 1472 works on other hosts
 */
#define MAX_BROADCAST_DATA_SIZE (1468 - DATA_PACKET_NONDATA_SIZE)
#define MAX_UNICAST_DATA_SIZE (UDP_SIZE - DATA_PACKET_NONDATA_SIZE)
#ifdef FreeBSD_BREAKAGE	/* FreeBSD packet fragmentation is broken */
#define MAX_MULTICAST_DATA_SIZE (1468 - DATA_PACKET_NONDATA_SIZE)
#else
#define MAX_MULTICAST_DATA_SIZE MAX_UNICAST_DATA_SIZE 
#endif

#define OPTION_CONFIRM_RECEIPT	0x01	/* subscribers should confirm msgs */
#define OPTION_RANDOM_WAIT	0x02	/* wait before requesting resend   */
#define OPTION_MORE1		0x04
#define OPTION_MORE2		0x08
#define OPTION_MORE3		0x10
#define OPTION_MORE4		0x20
#define OPTION_MORE5		0x40
#define OPTION_MORE6		0x80

typedef struct data_packet {
	group_id_t group_id;
	msg_id_t msg_id;		/* new id. for each message         */
	port_t	control_port;		/* where to send control messages   */
	time_t time;			/* timestamp of when I got perm.    */
	seq_t seq_number;		/* packet number		    */
	seq_t seq_total;		/* number of packets for this file  */
	u_long offset;			/* file offset for this packet      */
	long checksum;			/* only calculated on data field    */
	u_short data_size;		/* tells size of data array         */
	u_char packet_type;		/* MSG_SEND MSG_RESEND              */
	u_char options;			/* confirm_receipt or random_wait   */
	u_char authdata[16];            /* Authentication data              */
	data_t data[MAX_UNICAST_DATA_SIZE];	/* might not be all used    */
} data_packet_t;

#ifndef _POSIX_PATH_MAX		/* most systems define this in <limits.h> */
#define _POSIX_PATH_MAX  255    /* Number of bytes in a pathname */
#endif

/* NOTE: these 2 should fit within 1 MSG_SEND packet! */
#define MAX_FILENAME _POSIX_PATH_MAX /* minimum guaranteed by POSIX is 255 */
#define MAX_CMDNAME  _POSIX_PATH_MAX /* minimum guaranteed by POSIX is 255 */

/* begins the data portion of packet 1 */
typedef struct file_info {		/* this is filled in by client     */
	char name[MAX_FILENAME];	/* subscriber output file name     */
	char command[MAX_CMDNAME];	/* program subscriber should run   */
} file_info_t;
