/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * afdpjoin.c - join an AFDP group
 *
 * static char rcsid[] = "$Id: afdpjoin.c,v 1.66 1998/07/02 19:37:36 steve Exp $";
 */

#include <unistd.h>
#ifdef NEWSGI
#include <getopt.h>
#endif
#include "config.h"
#include "sockutils.h"
#include "afdp.h"
#include "afdplib.h"
#include "afdpjoin.h"
#ifdef ACL
#include "acl.h"
#endif

/* general global variables */
int	debug = 0;
group_id_t mygroup;
host_t  secretary_host = 0;
host_t  publisher_host = 0;
char *aclfile = NULL;	
char *safe_command = NULL;	
char *progname;
char *portname = "afdp";

boolean_t iam_secretary;
boolean_t iam_subscriber;
boolean_t daemonize = FALSE;
boolean_t reserved_ports = FALSE;
boolean_t opt_random_wait = TRUE;

/* sockets globals */
int 	search_sock;		/* UDP socket for find_secretary    */
int	admin_sock; 		/* TCP socket for join/leave        */
int	publish_sock; 		/* TCP socket for publish permit    */
int	group_data_sock; 	/* UDP socket for msg send/resend   */
port_t 	search_port;		/* port for find secretary          */
port_t 	admin_port; 		/* port for secretary<->subscribers */
port_t 	publish_port; 		/* port for client<->secretary      */
port_t 	group_data_port;	/* port for client->subscribers     */


int main(argc, argv)
int argc;
char *argv[];
{
	/* getopt variables first */
	int c;
	extern char *optarg;
	/* extern int optind, opterr; */
	int errflg = 0;
	int verflg = 0;

	group_id_t group = 1;
	search_port = htons(AFDP_SEARCH_PORT);	/* default value */
	progname = argv[0];
	setlog(progname, LOG_WARNING, LOG_STDERR, NULL, NULL);

	while ((c = getopt(argc,argv, "DVa:c:dg:ho:p:rs:IP:T:t:")) != -1) {
		switch (c) {
		/* standard afdp options first */
		case 'D':
			daemonize = TRUE;
			break;
		case 'V':
                    	verflg++;	/* print out version message */
			break;
	        case 'a':
#ifdef ACL
		        aclfile = optarg;
#else
			logmsg(LOG_ERR,"Access control lists not supported\n");
			exit(3);
#endif
			break;
		case 'c':
			safe_command = optarg;
			break;
		case 'd':
			debug++;
			break;
		case 'g':
			group = atol(optarg);
			break;
		case 'h':
                    	errflg++;	/* print out help message */
			break;
		case 'o':		/* afdp protocol options */
			if(strcmp(optarg, "nowait") == 0) {
				logmsg(LOG_DEBUG, "Option: random wait disabled");
				opt_random_wait = FALSE;
			}
			else
				logmsg(LOG_NOTICE,"BAD option: %s", optarg);
		case 'p':
			portname = optarg;
			/* logmsg(LOG_NOTICE,"portname is %s\n", portname); */
			break;
		case 'r':
			if(getuid()) {
			   logmsg(LOG_ERR,"must be root to use reserved ports\n");
			   exit(4);
			}
			reserved_ports = TRUE;
			portname = "afdp_res";
			search_port = htons(AFDP_RESERVED_PORT);
			/* logmsg(LOG_NOTICE,"reserved ports mode enabled\n"); */
			break;
		case 's':
			secretary_host = HostnameToAddr(optarg);
			break;
		/* options for afdpjoin, afdp subscriber & secretary */
		case 'I':
			threshold = INTERNET_THRESHOLD;
			break;
		case 'P':
			publisher_host = HostnameToAddr(optarg);
			break;
		case 'T':
			/* threshold is defined in secretary.c */
			threshold = atoi(optarg);
			break;
		case 't':
			/* multicast_ttl is defined in secretary.c */
			multicast_ttl = atoi(optarg);
			break;
               	default:
                    	errflg++;
               	}
          	if (errflg) 
			Usage();
	}

	if (verflg)
		Version();

	if (daemonize)
		setlog(progname, debug+LOG_WARNING, LOG_LOCAL7, LOG_PID, NULL);
	else
		setlog(progname, debug+LOG_WARNING, LOG_STDERR, NULL, NULL);

	if (debug)
		logmsg(LOG_NOTICE, "%s: debug level is %d\n", progname, debug);

	search_port = GetPort(portname, "udp", search_port);

	if (Check(group) == FAIL) {
		secretary_host = Hostaddr();
		BecomeSecretary();
		BecomeSubscriber(group);
	}
	else {
		iam_secretary = FALSE;
		if (Join(group) == FAIL)
			fatal("could not join group");
		BecomeSubscriber(group);
	}
	Receive();
	return(0);
}

/* check if secretary exists */
int Check(group)
group_id_t group;
{
	int status = FAIL;

	logmsg(LOG_INFO,"want to join group %lu\n", group);

	if (FindSecretary(group, &secretary_host, &admin_port,
				&publish_port, FALSE, search_port)) 
		status = SUCCESS;

	else {
		if (secretary_host && (secretary_host != Hostaddr())) {
			logmsg(LOG_NOTICE,"Couldn't find secretary on %s.\n",
                                HostaddrToName(secretary_host));
			exit(1);
		}
		logmsg(LOG_DEBUG,"tentatively become secretary for group %lu\n", group);
		if (FindSecretary(group, &secretary_host, 
			   &admin_port, &publish_port, TRUE, search_port)) 
			status = SUCCESS;
	}

	return (status);
}

/* if group does not exist, become secretary */
void BecomeSecretary()
{
#ifdef MULTICAST
	AddMember(secretary_host, MODE_MULTICAST);
#else
	AddMember(secretary_host, MODE_BROADCAST);
#endif

	iam_secretary = TRUE; 
	logmsg(LOG_NOTICE, "Secretary activated on %s\n", Hostname());
}

/* Become a subscriber */
void BecomeSubscriber(group)
group_id_t group;
{
	iam_subscriber = TRUE;	
	mygroup = group; /* now assign global variable my group id */
	logmsg(LOG_NOTICE, "Subscriber activated on %s\n", Hostname());
}

void Receive()
{
	struct sockaddr_in sin;
#ifdef MULTICAST
	struct ip_mreq mreq;
#endif
#ifdef SO_REUSEPORT
	int on = 1;
#endif

	/*
	 * For subscriber functions, first
	 * initialize port numbers - 0 if system assigned 
	 * then create sockets (with options) used by publisher 
	 * then bind sockets to ports if we plan to receive on them. 
	 */

	CreateSock (&group_data_sock, SOCK_DGRAM, 0);
	group_data_port = htons(0);
	BindSock (group_data_sock, &group_data_port, &sin);

#ifdef MULTICAST
	/* join the multicast group */
	mreq.imr_multiaddr.s_addr = MulticastAddr(mygroup);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	logmsg(LOG_DEBUG,"MulticastAddr is %s\n",
		HostaddrToName(mreq.imr_multiaddr.s_addr));

	if (setsockopt(group_data_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			(char *) &mreq, sizeof(mreq)) < 0) {
		fatal("setsockopt mreq: %m");
	}
#endif

	/* For secretary functions set up remaining ports. */

	if (iam_secretary) {
		admin_port = htons(0);
		publish_port = htons(0);

		CreateSock (&search_sock, SOCK_DGRAM, SO_REUSEADDR);
		CreateSock (&admin_sock, SOCK_STREAM, 0);
		CreateSock (&publish_sock, SOCK_STREAM, 0);

#ifdef SO_REUSEPORT
		if (setsockopt(search_sock,SOL_SOCKET,SO_REUSEPORT,(char *)&on,
					sizeof(on))< 0){
			fatal("setsockopt: SO_REUSEPORT: %m");
		}
#endif

		BindSock (search_sock, &search_port, &sin);
		BindSock (admin_sock, &admin_port, &sin);
		BindSock (publish_sock, &publish_port, &sin);

		/* listen on TCP sockets */
		ListenSock (admin_sock);
		ListenSock (publish_sock);
	}

	/*
	 * These assignments are needed to ensure that the sockets
	 * we wish to disregard do not contain values of legal sockets. 
	 * Otherwise, the code in GetPacket() will erroneously look
	 * at data on these sockets.
	 */

	else {
		search_sock = FD_SETSIZE;
		admin_sock = FD_SETSIZE;
		publish_sock = FD_SETSIZE;
	}

	/*
	 * Loop forever, listening for incoming packets. 
	 * We first deal with all possible incoming messages,
	 * then check for timeouts.
	 */
 
	while (1) {
		GetPacket();
		ClockTick();   /* in timer.c */
	}
}

void GetPacket()
{
	struct timeval timeout;
	fd_set r, w, e;
	int msgsock;
	int i;

	FD_ZERO(&r);
	FD_ZERO(&w);
	FD_ZERO(&e);

	/*
	 * If we only FD_SET the appropriate sockets relevant to our
	 * current status (whether we are secretary or subscriber)
	 * then we will never pay attention to packets which do
	 * not interest us.
	 */

	if (iam_secretary) {
		FD_SET(search_sock, &r);
		FD_SET(admin_sock, &r);
		FD_SET(publish_sock, &r);
	}
	if (iam_subscriber) {
		FD_SET(group_data_sock, &r);
	}

	if (!daemonize)
		FD_SET(fileno(stdin), &r); /* this allows run-time tuning */

 	timeout.tv_sec =  0;
	timeout.tv_usec = 10;

	if (select(getdtablesize(), &r, &w, &e, &timeout) > 0) {

		if (FD_ISSET(search_sock, &r)) {
			logmsg(LOG_DEBUG, "\tGetpacket: rec on search socket\n");
			HandleSearchPacket();
		}

		if (FD_ISSET(admin_sock, &r)) {
			logmsg(LOG_DEBUG, "\tGetpacket: rec on admin socket\n");
			msgsock = accept (admin_sock, 0, 0);
#ifdef ioerror_IDENT
			logmsg(LOG_NOTICE,"IDENT: admin request from %s@%s\n",
				PeerUser(msgsock),
				HostaddrToName(PeerHost(msgsock)));
#endif
			if (msgsock < 0) {
				fatal("accept failed");
			}
			else
				HandleAdminPacket(msgsock);

			close (msgsock);  /* tear down connection */
		}

		if (FD_ISSET(publish_sock, &r)) {
			logmsg(LOG_DEBUG, "\tGetpacket: rec on publish socket\n");
			msgsock = accept (publish_sock, 0, 0);
#ifdef ioerror_IDENT
			logmsg(LOG_NOTICE,"IDENT: publish request from %s@%s\n",
				PeerUser(msgsock),
				HostaddrToName(PeerHost(msgsock)));
#endif
			if (msgsock < 0) {
				fatal("accept failed");
			}
			else
				HandlePublishPacket(msgsock);

			close (msgsock);  /* tear down connection */
		}

		if (FD_ISSET(group_data_sock, &r)) {
			logmsg(LOG_DEBUG, "\tGetpacket: rec on data sock\n");
			HandleGroupDataPacket();
		}

		if (FD_ISSET(fileno(stdin), &r)) {
			char c;
			c = getchar();
			if (c =='+') {
				debug++;
				printf("debug level changed to %d\n", debug);
				setlog(progname, debug+LOG_WARNING, LOG_STDERR,
				NULL, NULL);
			}
			else if (c =='-') {
				debug--;
				printf("debug level changed to %d\n", debug);
				setlog(progname, debug+LOG_WARNING, LOG_STDERR,
				NULL, NULL);
			}
			else if (c =='q') {
				if (iam_subscriber == TRUE) {
					iam_subscriber = FALSE;
					close(group_data_sock);
					puts("subscription has been cancelled");
					if (iam_secretary == FALSE) {
						Leave();
						exit(0);
					}
					else if((i=IsMember(secretary_host))) {
						DelMember(i);
					}
				}
				else 
					puts("I am not a subscriber");
			}

			else if (c == 'Q') {
				if (iam_secretary == TRUE) {
					iam_secretary = FALSE;
					puts("quitting secretary duties");
					Shutdown();
					exit(0);
				}	
				else
					puts("do I look like a secretary?");
			}
		}
	}
}

/* this only works if all the hosts are on the same LAN right now */
void Shutdown ()
{
	data_packet_t	  shutdown;	/* will be sent to subscribers */
	int		  sock;

	logmsg(LOG_NOTICE, "Shutting down my group before quitting ...\n");

	/* send a broadcast (UDP) shutdown packets to all subscribers */
	/* it will be their job to send a TCP shutdown to all clients */
	logmsg(LOG_NOTICE, "\tShutting down all subscribers ...\n");
	shutdown.packet_type = SHUTDOWN;
	shutdown.group_id = htonl (mygroup);
	shutdown.data_size = htons(0);

	close(search_sock);
	close(admin_sock);
	close(publish_sock);
	CreateSock(&sock, SOCK_DGRAM, SO_BROADCAST);
	Broadcast(sock, group_data_port, &shutdown, DATA_PACKET_NONDATA_SIZE);
	Usleep(SECOND);
	Broadcast(sock, group_data_port, &shutdown, DATA_PACKET_NONDATA_SIZE);
	close(sock);
	logmsg(LOG_NOTICE, "\tAll done.\n");
}


void Usage () 
{
  printf("usage: %s [-DVdhr] [-a aclfile] [-g group] [-p port] [-s secretary_host]",
	progname);
  printf(" [-c safe_command]\n");
  printf("protocol options:\n");
  printf("\t-o nowait\n");
  printf("tuning options:\n");
  printf("\t-I [Internet mode]\n");
  printf("\t-T threshold\n");
  printf("\t-t multicast_ttl\n");
  exit (2);
}

void Version()
{
	printf("This is %s, version %s ", progname, AFDP_VERSION);
#ifdef MULTICAST
	printf("[with multicast support].\n");
#else
	printf("[without multicast support].\n");
#endif
	exit(2);
}
