/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * timer.c - handles timer processing
 *
 * static char rcsid[] = "$Id: timer.c,v 1.9 1998/07/02 19:37:36 steve Exp $";
 */

/*
 * 3 basic routines are provided:
 *
 * StartTimer (name, period, fcn, arg)
 * starts the named timer with a specified period (measured
 * in usec).  When the timer expires, the given function 
 * will be called. StartTimer correctly takes into account the 
 * elapsed time since the last update of the local clock. 
 *
 * CancelTimer (name)
 * removes the first occurrence of this timer from the timer 
 * queue. If there is more than one occurrence, too bad.
 *
 * ClockTick ()
 * updates the local clock time and calls the appropriate user
 * supplied function associated with each timer if any of them
 * have timed out during the last interval.  It is up to the 
 * user to restart the timer, if desired, following a timeout.
 */

#include "config.h"
#include "sockutils.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#define SCALE_SEC 1000000 	/* 1 sec = 10^6 usec */ 
#define SCALE_USEC 1 		/* 1 usec */

typedef struct entry {        
	u_short name;		/* timer name */
	long time;		/* time increment */
	int (*fcn)();		/* function to be called */
	u_long arg;		/* argument to be passed */
} entry_t;

typedef struct Queue {
	struct entry data;	/* queue entry */
	struct Queue *next;	/* pointer to next entry */
} Queue_t;

/* local global variables */
static Queue_t *t_head;		/* head of timer queue */
static struct timeval current_time;
static struct timezone tz;


void StartTimer (name, period, fcn, arg)
u_short name;			/* timer name */
long period;			/* timer period */
int (*fcn)();			/* function to call when timer expires */
u_long arg;			/* argument to be passed to function */
{
	long t_remain;		/* remaining time after q'th timeout */
	long delta;
	Queue_t *p;		/* previous queue element */
	Queue_t *q;		/* current queue element */
	Queue_t *t;		/* new queue element being inserted */
	struct timeval prev_time;
	struct timeval new_time;

	/*
	 * Have to take into account the amount of time that has
	 * already passed since our last ClockInc(), otherwise
	 * our new entry will not be correct.  But we don't want
 	 * to update the local clock here, since we're not going
	 * to act on it now. 
	 */

	prev_time = current_time;
	gettimeofday(&new_time, &tz);  	/* get time now */

	if (new_time.tv_usec >= prev_time.tv_usec)
		delta = (new_time.tv_sec - prev_time.tv_sec) * SCALE_SEC +
			(new_time.tv_usec - prev_time.tv_usec) * SCALE_USEC;
	else
		delta = (new_time.tv_sec - prev_time.tv_sec) * SCALE_SEC -
			(prev_time.tv_usec - new_time.tv_usec) * SCALE_USEC;

	t_remain = period + delta;	/* in time frame of previous ClockInc */
	p = NULL;
	q = t_head;

	while (q && t_remain > q->data.time) {
		t_remain -= q->data.time;
		p = q;
		q = q->next;
	}

	t = (struct Queue *) malloc (sizeof (struct Queue));
	t->data.name = name;
	t->data.time = t_remain;
	t->data.fcn = fcn;
	t->data.arg = arg;
	t->next = q;

	if (!p) 
		t_head = t;
	else 
		p->next = t;

	if (q) 
		q->data.time -= t_remain;
}

void CancelTimer (name)
u_short name;
{
	Queue_t *p;		/* previous queue element */
	Queue_t *q;		/* current queue element */

	p = NULL;
	q = t_head;

	while (q && q->data.name != name) {
		p = q;
		q = q->next;
	}

	if (q) {
		if (p) 
			p->next = q->next;
		else
			t_head = q->next;

		if (q->next)
			q->next->data.time += q->data.time;

		free (q);
	}
}

void ClockTick ()
{
	struct timeval prev_time;
	long delta;
	Queue_t *q;		/* current queue element */
	Queue_t *t;		/* temp queue element */

	prev_time = current_time;
	gettimeofday(&current_time, &tz);   	/* get time now */

	/* figure out how much time has elapsed since last call */

	if (current_time.tv_usec >= prev_time.tv_usec)
		delta = (current_time.tv_sec - prev_time.tv_sec) * SCALE_SEC +
			(current_time.tv_usec - prev_time.tv_usec) * SCALE_USEC;
	else
		delta = (current_time.tv_sec - prev_time.tv_sec) * SCALE_SEC -
			(prev_time.tv_usec - current_time.tv_usec) * SCALE_USEC;

	/*
	 * This gets really tricky.  The user function we activate
 	 * will likely call StartTimer to restart the timer, so
	 * we have to make sure that the new timer is inserted
 	 * at the right point in the queue with the correct
	 * period set.   To prevent the new timer being inserted
 	 * anywhere before the last timer which has now timed out,
	 * we first loop through the queue and set all the timed
	 * out timers to have zero remaining time, then go back
	 * and service the timeouts.
	 *
	 * For the first loop through the queue, 
	 * we want to catch all the timers that time out at the
	 * very same instant, not just the first one, so even if 
	 * delta goes to 0, we should keep checking while the
	 * remaining timers have zero time increments.
	 */

	q = t_head;
	while (q && (delta || !q->data.time)) {  
		if (q->data.time > delta) {	/* no timeout yet */
			q->data.time -= delta;
			delta = 0;
		}

		else {				/* a timer has expired */
			delta -= q->data.time;
			q->data.time = 0;
			q = q->next;
		}
	}

	/* 
	 * Now find all the timeouts and service them.
	 * As we do so, advance the pointer to the head
	 * head of the list so that this is always 
	 * correctly maintained.
	 */
	while (t_head && !t_head->data.time) {  
		(*t_head->data.fcn)(t_head->data.arg);	/* call function */
		t = t_head;
		t_head = t_head->next;
		free (t);
	}
}
